#include <csignal>
#include <cstring>
#include <sstream>
#include <string>
#include <thread>

#include "common/file_parser/helper_h264_parser.h"
#include "common/helper.h"
#include "common/opt_parser.h"
#include "common/sample_common.h"
#include "common/sample_connection_observer.h"
#include "common/sample_event.h"
#include "common/sample_local_user_observer.h"
#include "common/write_csvfile.h"
#include "common/log.h"

#define DEFAULT_SEND_MSG "start"
#define DEFAULT_MAX_DATASTREAM_WAIT_TIME 10000
#define DEFAULT_STATISTICS_COUNT (100)
#define DEFAULT_CONNECT_TIMEOUT_MS (3000)

static uint64_t send_datastream_T1 = 0;
static uint64_t receive_h264_T2 = 0;
static uint64_t round_count = 0;
static uint64_t sumSpendTime = 0;
static SampleEvent receive_ready_(true);

static std::unique_ptr<WriteCSVFileHandle> csv_file_handle = nullptr;
static int statistics_count_per_time = 0;

struct SampleOptions {
  std::string appId;
  std::string channelId;
  std::string userId;
  std::string remoteUserId;
  std::string data_stream_reliable = "true";
  std::string datastream_info = DEFAULT_SEND_MSG;
  std::string vos_ip;
  int statistics_count = DEFAULT_STATISTICS_COUNT;
};

static int waitUntilReceiveMessage(int waitMs) {
  return receive_ready_.Wait(waitMs);
}

class H264FrameReceiver : public agora::rtc::IVideoEncodedImageReceiver {
 public:
  H264FrameReceiver() {}

  bool OnEncodedVideoImageReceived(
      const uint8_t* imageBuffer, size_t length,
      const agora::rtc::EncodedVideoFrameInfo& videoEncodedFrameInfo) override;

 private:
  uint64_t frameNum = 0;
};

bool H264FrameReceiver::OnEncodedVideoImageReceived(
    const uint8_t* imageBuffer, size_t length,
    const agora::rtc::EncodedVideoFrameInfo& videoEncodedFrameInfo) {
  frameNum++;
  printf(
      "[ %s ]--> Now  Receive %luth Encode Video Image Data,  Image Data "
      "length = %d\n",
      getCurrentSystemTimeChrono().c_str(), frameNum, static_cast<int>(length));

  receive_h264_T2 = 0;
  receive_h264_T2 = now_ms_t();

  int spendTime = receive_h264_T2 - send_datastream_T1;
  if (spendTime <= 20) {
    std::this_thread::sleep_for(std::chrono::milliseconds(3000));
  } else {
    sumSpendTime += spendTime;
    round_count++;
    printf("send datastream then recv one h264 round_count:%lu,spendTime:%d\n",
           round_count, spendTime);

    delayTimeBaseInfo delay_info;
    delay_info.round_count_ = round_count;
    delay_info.spendTime_ = spendTime;
    delay_info.data_length = length;
    csv_file_handle->write_csvfile(delay_info);

    if (round_count % statistics_count_per_time == 0) {
      double averageSpendTime =
          sumSpendTime / (double)statistics_count_per_time;
      sumSpendTime = 0;
      printf(
          "the average spendTime from send datastream to receive h264 per "
          "%d times is %f \n",
          statistics_count_per_time, averageSpendTime);
    }
  }
  receive_ready_.Set();
  return true;
}

static bool exitFlag = false;
static void SignalHandler(int sigNo) {
  exitFlag = true;
  receive_ready_.Set();
}

int main(int argc, char* argv[]) {
  SampleOptions options;
  opt_parser optParser;

  optParser.add_long_opt("token", &options.appId,
                         "The token for authentication");
  optParser.add_long_opt("channelId", &options.channelId, "Channel Id");
  optParser.add_long_opt("userId", &options.userId, "User Id / default is 0");
  optParser.add_long_opt("remoteUserId", &options.remoteUserId,
                         "The remote user to receive stream from");
  optParser.add_long_opt(
      "reliable", &options.data_stream_reliable,
      "data stream transfer reliable (true/false)/ default reliable");
  optParser.add_long_opt("datastream_info", &options.datastream_info,
                         "The content you want to send datastream");
  optParser.add_long_opt("statistics_count", &options.statistics_count,
                         "statistics count per time");
  optParser.add_long_opt("vos_ip", &options.vos_ip,
                         "set the vos ip:port address ");

  if ((argc <= 1) || !optParser.parse_opts(argc, argv)) {
    std::ostringstream strStream;
    optParser.print_usage(argv[0], strStream);
    std::cout << strStream.str() << std::endl;
    return -1;
  }

  if (options.appId.empty()) {
    AG_LOG(ERROR, "Must provide appId!");
    return -1;
  }

  if (options.channelId.empty()) {
    AG_LOG(ERROR, "Must provide channelId!");
    return -1;
  }
  statistics_count_per_time = options.statistics_count;
  std::signal(SIGQUIT, SignalHandler);
  std::signal(SIGABRT, SignalHandler);
  std::signal(SIGINT, SignalHandler);

  // Create Agora service
  agora::base::IAgoraService* service = createAndInitAgoraService(false, true, true);
  if (!service) {
    AG_LOG(ERROR, "Failed to creating Agora service!");
    return -1;
  }
  // create connection
  agora::rtc::RtcConnectionConfiguration ccfg;
  ccfg.autoSubscribeAudio = false;
  ccfg.autoSubscribeVideo = false;
  ccfg.clientRoleType = agora::rtc::CLIENT_ROLE_BROADCASTER;
  ccfg.channelProfile = agora::CHANNEL_PROFILE_CLOUD_GAMING;

  agora::agora_refptr<agora::rtc::IRtcConnection> connection =
      service->createRtcConnection(ccfg);
  if (!connection) {
    AG_LOG(ERROR, "Failed to creating Agora connection!");
    return -1;
  }
  // Subcribe streams from all remote users or specific remote user
  agora::rtc::VideoSubscriptionOptions subscriptionOptions;
  subscriptionOptions.encodedFrameOnly = true;
  if (options.remoteUserId.empty()) {
    AG_LOG(INFO, "Subscribe streams from all remote users");
    connection->getLocalUser()->subscribeAllVideo(subscriptionOptions);
  } else {
    connection->getLocalUser()->subscribeVideo(options.remoteUserId.c_str(),
                                               subscriptionOptions);
  }

  // set vos server ip address
  agora::base::IAgoraParameter* agoraParameter =
      connection->getAgoraParameter();
  if (options.vos_ip.size() > 0) {
    std::string vos_ip_info = "\"" + options.vos_ip + "\"";
    std::string vos_list = "{\"rtc.vos_list\":[" + vos_ip_info + "]}";
    agoraParameter->setParameters(vos_list.c_str());
  }
  auto ConnObserver = std::make_shared<SampleConnectionObserver>();
  connection->registerObserver(ConnObserver.get());

  // Connect to Agora channel
  if (connection->connect(options.appId.c_str(), options.channelId.c_str(),
                          options.userId.c_str())) {
    AG_LOG(ERROR, "Failed to connect to Agora channel!");
    return -1;
  }

  // Create local user observer
  auto localUserObserver =
      std::make_shared<SampleLocalUserObserver>(connection->getLocalUser());

  // Register h264 frame receiver to receive video stream
  auto h264FrameReceiver = std::make_shared<H264FrameReceiver>();
  localUserObserver->setVideoEncodedImageReceiver(h264FrameReceiver.get());

  // save result to csv file
  std::unique_ptr<WriteCSVFileHandle> csv_handle_(new WriteCSVFileHandle(
      "./out_file/send_datastream_receive_h264_delay.csv"));
  csv_file_handle = std::move(csv_handle_);
  csv_file_handle->open();

  int send_stream_id = 66666;

  if (options.data_stream_reliable == "true" ||
      options.data_stream_reliable == "1") {
    connection->createDataStream(&send_stream_id, true, true, false);
    printf("------------------reliable----------------------------------\n");
  } else {
    connection->createDataStream(&send_stream_id, false, false, false);
    printf("------------------unreliable----------------------------------\n");
  }

  int msg_len = options.datastream_info.length();
  // Start receiving incoming media data
  AG_LOG(INFO, "Start receiving video data ...");

  // Periodically check exit flag
  while (!exitFlag) {
    int iret = waitUntilReceiveMessage(DEFAULT_MAX_DATASTREAM_WAIT_TIME);
    if (exitFlag) {
      continue;
    }
    if (iret < 0 && !exitFlag) {
      printf("loss datastream packet or loss h264 frame !!!!\n");
    }
    send_datastream_T1 = 0;
    send_datastream_T1 = now_ms_t();
    printf(
        "[ %s ]--> Now  Send a Data Stream Message,  Data stream Send: "
        "streamId = %d, length = %d\n",
        getCurrentSystemTimeChrono().c_str(), send_stream_id,
        static_cast<int>(msg_len));

    connection->sendStreamMessage(send_stream_id, options.datastream_info.data(),
                               msg_len);
  }

  // Unregister connection observer
  connection->unregisterObserver(ConnObserver.get());

  // Disconnect from Agora channel
  if (connection->disconnect()) {
    AG_LOG(ERROR, "Failed to disconnect from Agora channel!");
    return -1;
  }

  AG_LOG(INFO, "Disconnected from Agora channel successfully");
  ConnObserver.reset();
  h264FrameReceiver.reset();
  localUserObserver.reset();
  connection = nullptr;
  csv_file_handle->close();
  csv_file_handle.reset();
  return 0;
}
